@@ -45,6 +45,7 @@ static	int	varsub ARGS((Expand *xp, char *sp, char *word, int *stypep, int *slen
 static	int	comsub ARGS((Expand *xp, char *cp));
 static	char   *trimsub ARGS((char *str, char *pat, int how));
 static	void	tmpglob ARGS((char *cp, XPtrV *wp, int markdirs));
+static void globit(XString *, char **, char *, XPtrV *, int);
 static char	*maybe_expand_tilde ARGS((char *p, XString *dsp, char **dpp,
 					  int isassign));
 static	char   *tilde ARGS((char *acp));
@@ -153,10 +154,9 @@ typedef struct SubType {
 } SubType;
 
 void
-expand(cp, wp, f)
-	char *cp;		/* input word */
-	XPtrV *wp;	/* output words */
-	int f;			/* DO* flags */
+expand(const char *cp,	/* input word */
+    XPtrV *wp,		/* output words */
+    int f)		/* DO* flags */
 {
 	int UNINITIALIZED(c);
 	int type;	/* expansion type */
@@ -170,6 +170,7 @@ expand(cp, wp, f)
 	int UNINITIALIZED(newlines); /* For trailing newlines in COMSUB */
 	int saw_eq, tilde_ok;
 	int make_magic;
+	size_t len;
 
 	if (cp == NULL)
 		internal_errorf(1, "expand(NULL)");
@@ -266,33 +267,34 @@ expand(cp, wp, f)
 					}
 				}
 				continue;
-			  case OSUBST: /* ${{#}var{:}[=+-?#%]word} */
+			case OSUBST: /* ${{#}var{:}[=+-?#%]word} */
 			  /* format is:
 			   *   OSUBST [{x] plain-variable-part \0
 			   *     compiled-word-part CSUBST [}x]
-			   * This is were all syntax checking gets done...
+			   * This is where all syntax checking gets done...
 			   */
-			  {
-				char *varname = ++sp; /* skip the { or x (}) */
+			    {
+				const char *varname = ++sp; /* skip the { or x (}) */
 				int stype;
-				int slen;
+				int slen = 0;
 
-				sp = strchr(sp, '\0') + 1; /* skip variable */
+				sp = ((const char *)strchr(sp, '\0')) + 1; /* skip variable */
 				type = varsub(&x, varname, sp, &stype, &slen);
 				if (type < 0) {
-					char endc;
-					char *str, *end;
+					char *beg, *end, *str;
 
-					end = (char *) wdscan(sp, CSUBST);
+ unwind_substsyn:
+					sp = varname - 2; /* restore sp */
+					end = (beg = wdcopy(sp, ATEMP)) +
+					    (wdscan(sp, CSUBST) - sp);
 					/* ({) the } or x is already skipped */
-					endc = *end;
-					*end = EOS;
-					str = snptreef((char *) 0, 64, "%S",
-							varname - 1);
-					*end = endc;
+					if (end < wdscan(beg, EOS))
+						*end = EOS;
+					str = snptreef(NULL, 64, "%S", beg);
+					afree(beg, ATEMP);
 					errorf("%s: bad substitution", str);
 				}
-				if (f&DOBLANK)
+				if (f & DOBLANK)
 					doblank++;
 				tilde_ok = 0;
 				if (type == XBASE) {	/* expand? */
@@ -300,8 +302,8 @@ expand(cp, wp, f)
 						SubType *newst;
 
 						newst = (SubType *) alloc(
-							sizeof(SubType), ATEMP);
-						newst->next = (SubType *) 0;
+						    sizeof(SubType), ATEMP);
+						newst->next = NULL;
 						newst->prev = st;
 						st->next = newst;
 					}
@@ -390,7 +392,7 @@ expand(cp, wp, f)
 						doblank++;
 					st = st->prev;
 					continue;
-				  case '=':
+				case '=':
 					/* Restore our position and substitute
 					 * the value of st->var (may not be
 					 * the assigned value in the presence
@@ -408,49 +410,49 @@ expand(cp, wp, f)
 					 * fatal for special builtins (setstr
 					 * does readonly check).
 					 */
-					setstr(st->var, debunk(
-						(char *) alloc(strlen(dp) + 1,
-							ATEMP), dp),
-						KSH_UNWIND_ERROR);
+					len = strlen(dp) + 1;
+					setstr(st->var,
+					    debunk((char *) alloc(len, ATEMP),
+					    dp, len), KSH_UNWIND_ERROR);
 					x.str = str_val(st->var);
 					type = XSUB;
 					if (f&DOBLANK)
 						doblank++;
 					st = st->prev;
 					continue;
-				  case '?':
+				case '?':
 				    {
 					char *s = Xrestpos(ds, dp, st->base);
 
 					errorf("%s: %s", st->var->name,
-					    dp == s ? 
-					      "parameter null or not set"
-					    : (debunk(s, s), s));
+					    dp == s ?
+					    "parameter null or not set" :
+					    (debunk(s, s, strlen(s) + 1), s));
 				    }
 				}
 				st = st->prev;
 				type = XBASE;
 				continue;
 
-			  case OPAT: /* open pattern: *(foo|bar) */
+			case OPAT: /* open pattern: *(foo|bar) */
 				/* Next char is the type of pattern */
 				make_magic = 1;
 				c = *sp++ + 0x80;
 				break;
 
-			  case SPAT: /* pattern seperator (|) */
+			case SPAT: /* pattern separator (|) */
 				make_magic = 1;
 				c = '|';
 				break;
 
-			  case CPAT: /* close pattern */
+			case CPAT: /* close pattern */
 				make_magic = 1;
 				c = /*(*/ ')';
 				break;
 			}
 			break;
 
-		  case XNULLSUB:
+		case XNULLSUB:
 			/* Special case for "$@" (and "${foo[@]}") - no
 			 * word is generated if $# is 0 (unless there is
 			 * other stuff inside the quotes).
@@ -569,7 +571,7 @@ expand(cp, wp, f)
 				else if ((f & DOPAT) || !(fdo & DOMAGIC_))
 					XPput(*wp, p);
 				else
-					XPput(*wp, debunk(p, p));
+					XPput(*wp, debunk(p, p, strlen(p) + 1));
 				fdo = 0;
 				saw_eq = 0;
 				tilde_ok = (f & (DOTILDE|DOASNTILDE)) ? 1 : 0;
@@ -876,7 +878,7 @@ trimsub(char *str, char *pat, int how)
 	  case '#':		/* shortest at begining */
 		for (p = str; p <= end; p++) {
 			c = *p; *p = '\0';
-			if (gmatch(str, pat, FALSE)) {
+			if (gmatchx(str, pat, FALSE)) {
 				*p = c;
 				return p;
 			}
@@ -886,7 +888,7 @@ trimsub(char *str, char *pat, int how)
 	  case '#'|0x80:	/* longest match at begining */
 		for (p = end; p >= str; p--) {
 			c = *p; *p = '\0';
-			if (gmatch(str, pat, FALSE)) {
+			if (gmatchx(str, pat, FALSE)) {
 				*p = c;
 				return p;
 			}
@@ -895,13 +897,13 @@ trimsub(char *str, char *pat, int how)
 		break;
 	  case '%':		/* shortest match at end */
 		for (p = end; p >= str; p--) {
-			if (gmatch(p, pat, FALSE))
+			if (gmatchx(p, pat, FALSE))
 				return str_nsave(str, p - str, ATEMP);
 		}
 		break;
 	  case '%'|0x80:	/* longest match at end */
 		for (p = str; p <= end; p++) {
-			if (gmatch(p, pat, FALSE))
+			if (gmatchx(p, pat, FALSE))
 				return str_nsave(str, p - str, ATEMP);
 		}
 		break;
@@ -919,12 +921,17 @@ trimsub(char *str, char *pat, int how)
 static void
 tmpglob(char *cp, XPtrV *wp, int markdirs)
 {
-	if (glob_str(debunk(cp, cp), wp, markdirs) == 0)
-		XPput(*wp, cp);
+	int oldsize = XPsize(*wp);
+
+	if (glob_str(cp, wp, markdirs) == 0)
+		XPput(*wp, debunk(cp, cp, strlen(cp) + 1));
+	else
+		qsort(XPptrv(*wp) + oldsize, XPsize(*wp) - oldsize,
+		    sizeof (void *), xstrcmp);
 }
 
 #define GF_NONE		0
-#define GF_EXCHECK	BIT(0)		/* do existance check on file */
+#define GF_EXCHECK	BIT(0)		/* do existence check on file */
 #define GF_GLOBBED	BIT(1)		/* some globbing has been done */
 #define GF_MARKDIR	BIT(2)		/* add trailing / to directories */
 
@@ -934,36 +941,165 @@ tmpglob(char *cp, XPtrV *wp, int markdirs)
 int
 glob_str(char *cp, XPtrV *wp, int markdirs)
 {
-	unsigned int i;
-	glob_t buf;
+	int oldsize = XPsize(*wp);
+	XString xs;
+	char *xp;
 
-	glob(cp, markdirs ? GLOB_MARK : 0, NULL, &buf);
-	for(i = 0; i<buf.gl_pathc; i++) {
-		XPput(*wp, buf.gl_pathv[i]);
+	Xinit(xs, xp, 256, ATEMP);
+	globit(&xs, &xp, cp, wp, markdirs ? GF_MARKDIR : GF_NONE);
+	Xfree(xs, xp);
+
+	return XPsize(*wp) - oldsize;
+}
+
+static void
+globit(XString *xs,	/* dest string */
+    char **xpp,		/* ptr to dest end */
+    char *sp,		/* source path */
+    XPtrV *wp,		/* output list */
+    int check)		/* GF_* flags */
+{
+	char *np;		/* next source component */
+	char *xp = *xpp;
+	char *se;
+	char odirsep;
+
+	/* This to allow long expansions to be interrupted */
+	intrcheck();
+
+	if (sp == NULL) {	/* end of source path */
+		/* We only need to check if the file exists if a pattern
+		 * is followed by a non-pattern (eg, foo*x/bar; no check
+		 * is needed for foo* since the match must exist) or if
+		 * any patterns were expanded and the markdirs option is set.
+		 * Symlinks make things a bit tricky...
+		 */
+		if ((check & GF_EXCHECK) ||
+		    ((check & GF_MARKDIR) && (check & GF_GLOBBED))) {
+#define stat_check()	(stat_done ? stat_done : \
+			    (stat_done = stat(Xstring(*xs, xp), &statb) < 0 \
+				? -1 : 1))
+			struct stat lstatb, statb;
+			int stat_done = 0;	 /* -1: failed, 1 ok */
+
+			if (lstat(Xstring(*xs, xp), &lstatb) < 0)
+				return;
+			/* special case for systems which strip trailing
+			 * slashes from regular files (eg, /etc/passwd/).
+			 * SunOS 4.1.3 does this...
+			 */
+			if ((check & GF_EXCHECK) && xp > Xstring(*xs, xp) &&
+			    xp[-1] == '/' && !S_ISDIR(lstatb.st_mode) &&
+			    (!S_ISLNK(lstatb.st_mode) ||
+			    stat_check() < 0 || !S_ISDIR(statb.st_mode)))
+				return;
+			/* Possibly tack on a trailing / if there isn't already
+			 * one and if the file is a directory or a symlink to a
+			 * directory
+			 */
+			if (((check & GF_MARKDIR) && (check & GF_GLOBBED)) &&
+			    xp > Xstring(*xs, xp) && xp[-1] != '/' &&
+			    (S_ISDIR(lstatb.st_mode) ||
+			    (S_ISLNK(lstatb.st_mode) && stat_check() > 0 &&
+			    S_ISDIR(statb.st_mode)))) {
+				*xp++ = '/';
+				*xp = '\0';
+			}
+		}
+		XPput(*wp, str_nsave(Xstring(*xs, xp), Xlength(*xs, xp), ATEMP));
+		return;
 	}
 
-	return buf.gl_pathc;
+	if (xp > Xstring(*xs, xp))
+		*xp++ = '/';
+	while (*sp == '/') {
+		Xcheck(*xs, xp);
+		*xp++ = *sp++;
+	}
+	np = strchr(sp, '/');
+	if (np != NULL) {
+		se = np;
+		odirsep = *np;	/* don't assume '/', can be multiple kinds */
+		*np++ = '\0';
+	} else {
+		odirsep = '\0'; /* keep gcc quiet */
+		se = sp + strlen(sp);
+	}
+
+
+	/* Check if sp needs globbing - done to avoid pattern checks for strings
+	 * containing MAGIC characters, open [s without the matching close ],
+	 * etc. (otherwise opendir() will be called which may fail because the
+	 * directory isn't readable - if no globbing is needed, only execute
+	 * permission should be required (as per POSIX)).
+	 */
+	if (!has_globbing(sp, se)) {
+		XcheckN(*xs, xp, se - sp + 1);
+		debunk(xp, sp, Xnleft(*xs, xp));
+		xp += strlen(xp);
+		*xpp = xp;
+		globit(xs, xpp, np, wp, check);
+	} else {
+		DIR *dirp;
+		struct dirent *d;
+		char *name;
+		int len;
+		int prefix_len;
+
+		/* xp = *xpp;	   copy_non_glob() may have re-alloc'd xs */
+		*xp = '\0';
+		prefix_len = Xlength(*xs, xp);
+		dirp = opendir(prefix_len ? Xstring(*xs, xp) : ".");
+		if (dirp == NULL)
+			goto Nodir;
+		while ((d = readdir(dirp)) != NULL) {
+			name = d->d_name;
+			if (name[0] == '.' &&
+			    (name[1] == 0 || (name[1] == '.' && name[2] == 0)))
+				continue; /* always ignore . and .. */
+			if ((*name == '.' && *sp != '.')
+			    || !gmatchx(name, sp, TRUE))
+				continue;
+
+			len = strlen(d->d_name) + 1;
+			XcheckN(*xs, xp, len);
+			memcpy(xp, name, len);
+			*xpp = xp + len - 1;
+			globit(xs, xpp, np, wp,
+				(check & GF_MARKDIR) | GF_GLOBBED
+				| (np ? GF_EXCHECK : GF_NONE));
+			xp = Xstring(*xs, xp) + prefix_len;
+		}
+		closedir(dirp);
+ Nodir:
+		;
+	}
+
+	if (np != NULL)
+		*--np = odirsep;
 }
 
 /* remove MAGIC from string */
 char *
-debunk(dp, sp)
-	char *dp;
-	const char *sp;
+debunk(char *dp, const char *sp, size_t dlen)
 {
-	char *d, *s;
+	char *d;
+	const char *s;
 
-	if ((s = strchr(sp, MAGIC))) {
+	if ((s = ((const char *)strchr(sp, MAGIC)))) {
+		if (s - sp >= (ssize_t)dlen)
+			return dp;
 		memcpy(dp, sp, s - sp);
-		for (d = dp + (s - sp); *s; s++)
-			if (!ISMAGIC(*s) || !(*++s & 0x80)
-			    || !strchr("*+?@! ", *s & 0x7f))
+		for (d = dp + (s - sp); *s && (d - dp < (ssize_t)dlen); s++)
+			if (!ISMAGIC(*s) || !(*++s & 0x80) ||
+			    !strchr("*+?@! ", *s & 0x7f))
 				*d++ = *s;
 			else {
 				/* extended pattern operators: *+?@! */
 				if ((*s & 0x7f) != ' ')
 					*d++ = *s & 0x7f;
-				*d++ = '(';
+				if (d - dp < (ssize_t)dlen)
+					*d++ = '(';
 			}
 		*d = '\0';
 	} else if (dp != sp)
