@@ -387,7 +387,7 @@ execute(t, flags)
 		for (t = t->left; t != NULL && t->type == TPAT; t = t->right)
 		    for (ap = t->vars; *ap; ap++)
 			if ((s = evalstr(*ap, DOTILDE|DOPAT))
-			    && gmatch(cp, s, FALSE))
+			    && gmatchx(cp, s, FALSE))
 				goto Found;
 		break;
 	  Found:
@@ -885,7 +885,7 @@ findfunc(const char *name, int create)
 		if (tp)
 			break;
 		if (!l->next && create) {
-			tp = transitional_tenter(&l->funs.root, name);
+			tp = transitional_tenter(&l->funs.root, name, APERM);
 			tp->flag = DEFINED;
 			tp->type = CFUNC;
 			tp->val.t = (struct op *) 0;
@@ -953,7 +953,7 @@ builtin(name, func, flags)
 
 	/* see if any flags should be set for this builtin */
 
-	tp = transitional_tenter(&builtins.root, name);
+	tp = transitional_tenter(&builtins.root, name, APERM);
 	tp->flag = DEFINED | flags;
 	tp->type = CSHELL;
 	tp->val.f = func;
