@@ -7,23 +7,23 @@
 #ifdef HAVE_LIMITS_H
 # include <limits.h>
 #endif
-#include <fnmatch.h>
 
 #ifndef UCHAR_MAX
 # define UCHAR_MAX	0xFF
 #endif
 
 short ctypes [UCHAR_MAX+1];	/* type bits for unsigned char */
+static int do_gmatch(const unsigned char *, const unsigned char *,
+		        const unsigned char *, const unsigned char *);
+static const unsigned char *cclass(const unsigned char *, int);
 
 /*
  * Fast character classes
  */
 void
-setctypes(s, t)
-	register const char *s;
-	register int t;
+setctypes(const char *s, int t)
 {
-	register int i;
+	unsigned i;
 
 	if (t & C_IFS) {
 		for (i = 0; i < UCHAR_MAX+1; i++)
@@ -54,7 +54,7 @@ str_save(s, ap)
 }
 
 /* Allocate a string of size n+1 and copy upto n characters from the possibly
- * null terminated string s into it.  Always returns a null terminated string
+ * NUL terminated string s into it.  Always returns a NUL terminated string
  * (unless n < 0).
  */
 char *
@@ -136,9 +136,9 @@ const struct posh_option options[] = {
 #endif
 	{ "xtrace",	'x',		OF_ANY },
 	/* Anonymous flags: used internally by shell only
-	 * (not visable to user)
+	 * (not visible to user)
 	 */
-	{ (char *) 0,	0,		OF_INTERNAL }, /* FTALKING_I */
+	{ NULL,		 0,	   OF_INTERNAL }, /* FTALKING_I */
 };
 
 /*
@@ -433,12 +433,9 @@ bi_getn(as, ai)
  */
 
 int
-gmatch(s, p, isfile)
-	const char *s, *p;
-	int isfile;
+gmatchx(const char *s, const char *p, int isfile)
 {
 	const char *se, *pe;
-	char *debunked;
 
 	if (s == NULL || p == NULL)
 		return 0;
@@ -448,23 +445,22 @@ gmatch(s, p, isfile)
 	 * the pattern.  If check fails, just to a strcmp().
 	 */
 	if (!isfile && !has_globbing(p, pe)) {
-		unsigned int len = pe - p + 1;
+		size_t len = pe - p + 1;
 		char tbuf[64];
-		char *t = len <= sizeof(tbuf) ? tbuf
-				: (char *) alloc(len, ATEMP);
-		debunk(t, p);
+		char *t = len <= sizeof(tbuf) ? tbuf :
+		    (char *) alloc(len, ATEMP);
+		debunk(t, p, len);
 		return !strcmp(t, s);
 	}
-	debunked = alloc(strlen(p)+1, ATEMP);
-	debunk(debunked, p);
-	return !fnmatch(debunked, s, isfile ? FNM_PATHNAME : 0);
+	return do_gmatch((const unsigned char *) s, (const unsigned char *) se,
+	    (const unsigned char *) p, (const unsigned char *) pe);
 }
 
 /* Returns if p is a syntacticly correct globbing pattern, false
  * if it contains no pattern characters or if there is a syntax error.
  * Syntax errors are:
  *	- [ with no closing ]
- *	- imballenced $(...) expression
+ *	- imbalanced $(...) expression
  *	- [...] and *(...) not nested (eg, [a$(b|]c), *(a[b|c]d))
  */
 /*XXX
@@ -477,8 +473,7 @@ gmatch(s, p, isfile)
 - return ?
 */
 int
-has_globbing(xp, xpe)
-	const char *xp, *xpe;
+has_globbing(const char *xp, const char *xpe)
 {
 	const unsigned char *p = (const unsigned char *) xp;
 	const unsigned char *pe = (const unsigned char *) xpe;
@@ -530,35 +525,202 @@ has_globbing(xp, xpe)
 	return saw_glob && !in_bracket && !nest;
 }
 
+/* Function must return either 0 or 1 (assumed by code for 0x80|'!') */
+static int
+do_gmatch(const unsigned char *s, const unsigned char *se,
+    const unsigned char *p, const unsigned char *pe)
+{
+	int sc, pc;
+	const unsigned char *prest, *psub, *pnext;
+	const unsigned char *srest;
+
+	if (s == NULL || p == NULL)
+		return 0;
+	while (p < pe) {
+		pc = *p++;
+		sc = s < se ? *s : '\0';
+		s++;
+		if (!ISMAGIC(pc)) {
+			if (sc != pc)
+				return 0;
+			continue;
+		}
+		switch (*p++) {
+		case '[':
+			if (sc == 0 || (p = cclass(p, sc)) == NULL)
+				return 0;
+			break;
+
+		case '?':
+			if (sc == 0)
+				return 0;
+			break;
+
+		case '*':
+			if (p == pe)
+				return 1;
+			s--;
+			do {
+				if (do_gmatch(s, se, p, pe))
+					return 1;
+			} while (s++ < se);
+			return 0;
+
+		  /*
+		   * [*+?@!](pattern|pattern|..)
+		   * This is also needed for ${..%..}, etc.
+		   */
+		case 0x80|'+': /* matches one or more times */
+		case 0x80|'*': /* matches zero or more times */
+			if (!(prest = pat_scan(p, pe, 0)))
+				return 0;
+			s--;
+			/* take care of zero matches */
+			if (p[-1] == (0x80 | '*') &&
+			    do_gmatch(s, se, prest, pe))
+				return 1;
+			for (psub = p; ; psub = pnext) {
+				pnext = pat_scan(psub, pe, 1);
+				for (srest = s; srest <= se; srest++) {
+					if (do_gmatch(s, srest, psub, pnext - 2) &&
+					    (do_gmatch(srest, se, prest, pe) ||
+					    (s != srest && do_gmatch(srest,
+					    se, p - 2, pe))))
+						return 1;
+				}
+				if (pnext == prest)
+					break;
+			}
+			return 0;
+
+		case 0x80|'?': /* matches zero or once */
+		case 0x80|'@': /* matches one of the patterns */
+		case 0x80|' ': /* simile for @ */
+			if (!(prest = pat_scan(p, pe, 0)))
+				return 0;
+			s--;
+			/* Take care of zero matches */
+			if (p[-1] == (0x80 | '?') &&
+			    do_gmatch(s, se, prest, pe))
+				return 1;
+			for (psub = p; ; psub = pnext) {
+				pnext = pat_scan(psub, pe, 1);
+				srest = prest == pe ? se : s;
+				for (; srest <= se; srest++) {
+					if (do_gmatch(s, srest, psub, pnext - 2) &&
+					    do_gmatch(srest, se, prest, pe))
+						return 1;
+				}
+				if (pnext == prest)
+					break;
+			}
+			return 0;
+
+		case 0x80|'!': /* matches none of the patterns */
+			if (!(prest = pat_scan(p, pe, 0)))
+				return 0;
+			s--;
+			for (srest = s; srest <= se; srest++) {
+				int matched = 0;
+
+				for (psub = p; ; psub = pnext) {
+					pnext = pat_scan(psub, pe, 1);
+					if (do_gmatch(s, srest, psub,
+					    pnext - 2)) {
+						matched = 1;
+						break;
+					}
+					if (pnext == prest)
+						break;
+				}
+				if (!matched &&
+				    do_gmatch(srest, se, prest, pe))
+					return 1;
+			}
+			return 0;
+
+		default:
+			if (sc != p[-1])
+				return 0;
+			break;
+		}
+	}
+	return s == se;
+}
+
+static const unsigned char *
+cclass(const unsigned char *p, int sub)
+{
+	int c, d, not, found = 0;
+	const unsigned char *orig_p = p;
+
+	if ((not = (ISMAGIC(*p) && *++p == NOT)))
+		p++;
+	do {
+		c = *p++;
+		if (ISMAGIC(c)) {
+			c = *p++;
+			if ((c & 0x80) && !ISMAGIC(c)) {
+				c &= 0x7f;/* extended pattern matching: *+?@! */
+				/* XXX the ( char isn't handled as part of [] */
+				if (c == ' ') /* simile for @: plain (..) */
+					c = '(' /*)*/;
+			}
+		}
+		if (c == '\0')
+			/* No closing ] - act as if the opening [ was quoted */
+			return sub == '[' ? orig_p : NULL;
+		if (ISMAGIC(p[0]) && p[1] == '-' &&
+		    (!ISMAGIC(p[2]) || p[3] != ']')) {
+			p += 2; /* MAGIC- */
+			d = *p++;
+			if (ISMAGIC(d)) {
+				d = *p++;
+				if ((d & 0x80) && !ISMAGIC(d))
+					d &= 0x7f;
+			}
+			/* POSIX says this is an invalid expression */
+			if (c > d)
+				return NULL;
+		} else
+			d = c;
+		if (c == sub || (c <= sub && sub <= d))
+			found = 1;
+	} while (!(ISMAGIC(p[0]) && p[1] == ']'));
+
+	return (found != not) ? p+2 : NULL;
+}
+
 /* Look for next ) or | (if match_sep) in *(foo|bar) pattern */
 const unsigned char *
-pat_scan(p, pe, match_sep)
-	const unsigned char *p;
-	const unsigned char *pe;
-	int match_sep;
+pat_scan(const unsigned char *p, const unsigned char *pe, int match_sep)
 {
 	int nest = 0;
 
 	for (; p < pe; p++) {
 		if (!ISMAGIC(*p))
 			continue;
-		if ((*++p == /*(*/ ')' && nest-- == 0)
-		    || (*p == '|' && match_sep && nest == 0))
+		if ((*++p == /*(*/ ')' && nest-- == 0) ||
+		    (*p == '|' && match_sep && nest == 0))
 			return ++p;
 		if ((*p & 0x80) && strchr("*+?@! ", *p & 0x7f))
 			nest++;
 	}
-	return (const unsigned char *) 0;
+	return NULL;
 }
 
-/* Initialize a Getopt structure */
+int
+xstrcmp(const void *p1, const void *p2)
+{
+	return (strcmp(*(char * const *)p1, *(char * const *)p2));
+}
+
+/* Initialise a Getopt structure */
 void
-ksh_getopt_reset(go, flags)
-	LameGetopt *go;
-	int flags;
+ksh_getopt_reset(LameGetopt *go, int flags)
 {
 	go->optind = 1;
-	go->optarg = (char *) 0;
+	go->optarg = NULL;
 	go->p = 0;
 	go->flags = flags;
 	go->info = 0;
